#!/usr/local/lib python2.7

import utils2D
import sys
import numpy as np
import scipy.optimize as opt


def main(fovy, amp, centershift, bw, cycle, input_ksp, output_psf):
    '''
    Main function for 2D wave-encoding PSF auto-calibration.
    Input kspace with shape [x, y, 1, coil].
    Writes out a cfl file of the PSF in shape [x, y].

    Reference:
    Chen F, Taviani V, Tamir JI, Cheng JY, Zhang T, Song Q, Hargreaves BA, 
    Pauly JM, Vasanawala SS. Self‐Calibrating Wave‐Encoded Variable‐Density 
    Single‐Shot Fast Spin Echo Imaging. Journal of Magnetic Resonance Imaging. 
    2018 Apr;47(4):954-66.

    Input --
        The first 5 parameters are related to the wave encoding waveforms.
        fov - 0.288 m.
        amp - 1.10.
        centershift - iso center.
        bw - 125.
        cycle - 5.
        input_ksp - filename of input kspace cfl file.
        output_psf - filename of output psf cfl file. 

    NOTE --
        Due to different waveform designs, the PSF generation function
        in utils2D.py (PsfGen, PsfGenHelper, and WaveformDesign)
        may need to be replaced with the actual design of wave encoding
        waveforms, as well as the parameters.
    '''
    data = utils2D.readcfl(input_ksp)
    data = data[:, :, 0, :]
    skx, sky, sc = data.shape
    print("Input Data shape: x %d y %d c %d" % (skx, sky, sc))

    # Input parameters.
    fovy = float(fovy)
    cycle = float(cycle)
    bw = float(bw)
    init_amp = float(amp)
    init_phase = 0.1  # needs to be non-zero.
    init_centershift = float(centershift)

    def PsfCalib_func(argin):
        var_amp, var_phase, var_centershift = argin
        psf, _ = utils2D.PsfGen(
            skx, sky, fovy, var_amp, cycle, bw, var_phase, var_centershift)
        im_tmp_zp = utils2D.ifft3cw(data, psf)
        current_entropy = 1.0e5 * utils2D.gradEntropy(im_tmp_zp)
        print(">>> Calibrating PSF... [amp %f phase %f centershift %f Entropy %f]..." %
              (var_amp, var_phase, var_centershift, current_entropy))
        return -current_entropy

    init = np.asarray([init_amp, init_phase, init_centershift])
    print("Input initialization: amp %f phase %f centershift %f" %
          (init_amp, init_phase, init_centershift))

    res = opt.fmin(PsfCalib_func, init)
    psf, _ = utils2D.PsfGen(
        skx, sky, fovy, res[0], cycle, bw, res[1], res[2])

    utils2D.writecfl(output_psf, psf)

    return 0


if __name__ == '__main__':

    exit(main(sys.argv[1], sys.argv[2], sys.argv[3], sys.argv[4],
              sys.argv[5], sys.argv[6], sys.argv[7]))
