import numpy as np
import cmath

import scipy.io as sio
import scipy.linalg
import scipy.optimize as opt
from scipy import interpolate


def readcfl(name):
    # Copyright 2013-2019. The Regents of the University of California.
    # All rights reserved. Use of this source code is governed by
    # a BSD-style license which can be found in the LICENSE file.
    #
    # Authors:
    # 2013 Martin Uecker <uecker@eecs.berkeley.edu>
    # 2015 Jonathan Tamir <jtamir@eecs.berkeley.edu>

    # get dims from .hdr
    h = open(name + ".hdr", "r")
    h.readline()  # skip
    l = h.readline()
    h.close()
    dims = [int(i) for i in l.split()]

    # remove singleton dimensions from the end
    n = np.prod(dims)
    dims_prod = np.cumprod(dims)
    dims = dims[:np.searchsorted(dims_prod, n) + 1]

    # load data and reshape into dims
    d = open(name + ".cfl", "r")
    a = np.fromfile(d, dtype=np.complex64, count=n)
    d.close()
    return a.reshape(dims, order='F')  # column-major


def writecfl(name, array):
    # Copyright 2013-2019. The Regents of the University of California.
    # All rights reserved. Use of this source code is governed by
    # a BSD-style license which can be found in the LICENSE file.
    #
    # Authors:
    # 2013 Martin Uecker <uecker@eecs.berkeley.edu>
    # 2015 Jonathan Tamir <jtamir@eecs.berkeley.edu>
    h = open(name + ".hdr", "w")
    h.write('# Dimensions\n')
    for i in (array.shape):
        h.write("%d " % i)
    h.write('\n')
    h.close()
    d = open(name + ".cfl", "w")
    array.T.astype(np.complex64).tofile(d)  # tranpose for column-major order
    d.close()


def ifftc(data, dim):
    tmp = np.fft.ifftshift(data, axes=dim)
    tmp = np.fft.ifft(tmp, axis=dim)
    tmp = np.fft.fftshift(tmp, axes=dim) * (data.shape[dim]**0.5)
    return tmp


def fftc(data, dim):
    tmp = np.fft.ifftshift(data, axes=dim)
    tmp = np.fft.fft(tmp, axis=dim)
    tmp = np.fft.fftshift(tmp, axes=dim) * 1.0 / (data.shape[dim]**0.5)
    return tmp


def fft2c(data):
    # data 3D, the last dim is coil channels
    factor = 1.0 / (data.shape[0] * data.shape[1])**0.5
    tmp = np.zeros(data.shape) * 0j
    for ic in range(data.shape[2]):
        tmp[:, :, ic] = np.fft.ifftshift(data[:, :, ic])
        tmp[:, :, ic] = np.fft.fft2(tmp[:, :, ic])
        tmp[:, :, ic] = np.fft.fftshift(tmp[:, :, ic]) * factor
    return tmp


def ifft2c_wave(data, psf):
    # data 2D psf 2D.
    tmp = ifftc(data, 1) * np.conj(psf)
    tmp = ifftc(tmp, 0)
    return tmp


def ifft3c_wave(data, psf):
    # data 3D psf 2D.
    tmp = ifftc(data, 1) * np.transpose(np.tile(np.conj(psf),
                                                (data.shape[2], 1, 1)), (1, 2, 0))
    tmp = ifftc(tmp, 0)
    return tmp


def wave2cart(data, psf):
    # data 3D psf 2D.
    tmp = ifftc(data, 1) * np.transpose(np.tile(np.conj(psf),
                                                (data.shape[2], 1, 1)), (1, 2, 0))
    tmp = fftc(tmp, 1)
    return tmp


def ifft3cw(data, psf):
    # psf is 2D, data is 3D.
    sc = data.shape[2]
    res = np.zeros(data.shape[0:2])
    for ic in range(0, sc):
        res = res + np.abs(ifft2c_wave(data[:, :, ic], psf))**2
    res = res**0.5
    return res


def gradEntropy(imdata):
    # imdata is 2D.
    imGx, imGy = np.gradient(imdata)
    imG = np.abs(imGx) + np.abs(imGy)
    out2 = (imG.flatten(1) / np.sum(imG.flatten(1)))**2
    out2 = np.sum(out2)
    return out2


def WaveformDesign(skx, bw, cycle, choice):
    '''
    Waveform generation from input parameters.
    Currently only choice = 1 supported.
    Returns key timing points.
    Can be replaced with custom waveform generation functions.
    '''
    if choice == 1:
        dt = 500.0 / bw
        pw_cycle_pre = skx * dt / ((2 * cycle - 1) * 1.4142 + 2)
        pw_cycle_pre = np.around(pw_cycle_pre / dt + 0.2)
        pw_cycle_half = skx * dt / ((2 * cycle - 1) * 1.4142 + 2) * 1.4142
        pw_cycle_half = np.around(pw_cycle_half / dt + 0.2)
        period = 2 * pw_cycle_half
        startpt = int((pw_cycle_pre) + 1)
        endpt = int((pw_cycle_pre) + pw_cycle_half * (2 * cycle - 1))
    return (period, startpt, endpt, pw_cycle_pre)


def PsfGenHelper(zsol1, zsol2, zsol3, zsol4, skx, sky, bw, cycle, choice=1):
    '''
    PSF generation based on waveform timing and some other paramters.
    Returns PSF and a 1D sample waveform parameters.
    Can be replaced with custom functions.
    '''
    per, startpt, endpt, pw_cycle_pre = WaveformDesign(skx, bw, cycle, 1)

    def IntWave(wv):
        tmpwv = np.cumsum(wv)
        tmpwv = tmpwv / np.max(np.abs(tmpwv))  # To the same scale
        return tmpwv

    def fit(a, x):
        return a * (np.sin(2 * cmath.pi * x / per))

    def fit2(a, x):
        return 0.707 * a * (np.sin(2 * cmath.pi * x / (pw_cycle_pre * 2) + cmath.pi))

    xlist = np.asarray(range(1, int(endpt - startpt + 1) + 1))  # vector
    linCoeffzdelay = np.zeros(10)
    linCoeffzpre = fit2(1, np.asarray(range(1, startpt)))
    linCoeffz = fit(1, xlist)
    linCoeffzpost = fit2(1, np.asarray(range(1, skx - endpt + 1)))
    linCoeffzend = np.zeros(10)

    tmp = np.concatenate((linCoeffzdelay, linCoeffzpre))
    tmp = np.concatenate((tmp, linCoeffz))
    tmp = np.concatenate((tmp, linCoeffzpost))
    tmp = np.concatenate((tmp, linCoeffzend))

    linCoeffzInt = IntWave(tmp)
    linCoeffzInt = zsol1 * linCoeffzInt + zsol3

    interp_func = interpolate.interp1d(np.asarray(
        range(-10, int(skx) + 10)) + zsol2, linCoeffzInt)
    linCoeffz = interp_func(np.asarray(np.arange(0, int(skx))))

    psfz = np.zeros((skx, sky))
    for ix in range(0, skx):
        psfz[ix, ] = linCoeffz[ix] * (np.asarray(range(1, sky + 1)) - zsol4)
    centershifty = zsol4 - 0.5 - sky / 2.0

    psf = np.zeros((skx, sky))
    psf = np.exp(1j * psfz)
    params = [skx, sky, 1, bw, centershifty, 0]
    params = np.concatenate((np.asarray(params), linCoeffz))
    return (psf, params)


def PsfGen(skx, sky, fovy, xopamp, cycle, bw, xopphase, xcentershifty):
    '''
    Main PSF generation function.
    Returns PSF and a 1D sample waveform parameters.
    Can be replaced with custom function.
    '''
    resy = fovy / float(sky)
    facy_builtin = 1.0323
    tsy = 1.0 / xopphase
    Gamp = xopamp * 1e-2
    ROcycle = cycle
    ROskx = skx
    RObw = bw * 1e3
    T = float(ROskx) / 2.0 / RObw / ROcycle

    zsol1 = 42.576e6 * T * Gamp * resy / facy_builtin   # Amplitude.
    zsol2 = tsy   # Inverse phase shift.
    zsol3 = 0   # DC shift.
    zsol4 = xcentershifty  # Iso center.
    
    return PsfGenHelper(zsol1, zsol2, zsol3, zsol4, skx, sky, bw, cycle, 1)
